/*****************************************************************************
* Product: QDPP example
* Last Updated for Version: 4.0.00
* Date of the Last Update:  Apr 07, 2008
*
*                    Q u a n t u m     L e a P s
*                    ---------------------------
*                    innovating embedded systems
*
* Copyright (C) 2002-2008 Quantum Leaps, LLC. All rights reserved.
*
* This software may be distributed and modified under the terms of the GNU
* General Public License version 2 (GPL) as published by the Free Software
* Foundation and appearing in the file GPL.TXT included in the packaging of
* this file. Please note that GPL Section 2[b] requires that all works based
* on this software must also be made publicly available under the terms of
* the GPL ("Copyleft").
*
* Alternatively, this software may be distributed and modified under the
* terms of Quantum Leaps commercial licenses, which expressly supersede
* the GPL and are specifically designed for licensees interested in
* retaining the proprietary status of their code.
*
* Contact information:
* Quantum Leaps Web site:  http://www.quantum-leaps.com
* e-mail:                  info@quantum-leaps.com
*****************************************************************************/
#include <boost/asio/io_service.hpp>
#include <boost/asio/steady_timer.hpp>
#include <boost/chrono.hpp>

typedef boost::system::error_code error_code;

extern "C" {
#include "config.h"

#include "qp_port.h"
#include "bsp.h"
#include "qf.h"
#include "qvanilla.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>                            /* for memcpy() and memset() */

#ifdef Q_SPY
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <errno.h>

    static int l_sock = -1;

    enum AppRecords {                 /* application-specific trace records */
        PHILO_STAT = QS_USER
    };
#endif

#include "util.h"

Q_DEFINE_THIS_FILE

/* Local objects -----------------------------------------------------------*/
static uint32_t l_delay __attribute__((unused));       /* limit for the loop counter in busyDelay() */

namespace {
class Bsp
{
public:
    Bsp(boost::asio::io_service& ioService)
        : m_ioService(ioService)
        , m_tickTimer(ioService)
        , m_isPollPosted(false)
    {
        asyncWaitTickTimer(boost::chrono::steady_clock::now());
    }

    void notify()
    {
        if (m_isPollPosted) {
            return;
        }
        m_isPollPosted = true;
        m_ioService.post([=]()
        {
            m_isPollPosted = false;
            QF_poll();
        });
    }

    void stop()
    {
        m_ioService.stop();
    }

    boost::asio::io_service& ioService() { return m_ioService; }
private:
    boost::asio::io_service& m_ioService;
    boost::asio::basic_waitable_timer<boost::chrono::steady_clock> m_tickTimer;
    bool m_isPollPosted;


    void asyncWaitTickTimer(boost::chrono::time_point<boost::chrono::steady_clock> lastTime)
    {
        m_tickTimer.expires_at(lastTime);
        m_tickTimer.async_wait([=](const error_code& e)
        {
            if (e) {
                return; //timer was aborted
            }

            QF_tick();
            asyncWaitTickTimer(lastTime + boost::chrono::milliseconds(BSP_MS_PER_TICKS));
        });
    }
};

//singleton instance created in BSP_init and deleted in BSP_destroy
Bsp *g_bsp;

} //namespace

/*..........................................................................*/
#ifdef Q_SPY
void BSP_init(int argc, char *argv[]) {
    char const *hostAndPort = "localhost:6601";

    if (argc > 1) {
        l_delay = atol(argv[1]);         /* set the delay from the argument */
    }
    if (argc > 2) {                                      /* port specified? */
        hostAndPort = argv[2];
    }
    if (!QS_INIT(hostAndPort)) {
        printf("\nUnable to open QS socket\n");
        QF_stop();
    }
    printf("Dining Philosopher Problem example"
           "\nQEP %s\nQF  %s\n"
           "Press ESC to quit...\n",
           QEP_getVersion(),
           QF_getVersion());

    static_assert(false, "NOT MAINTAINED CODE !!!");
}
#endif

void BSP_init(boost::asio::io_service &ioService)
{
    g_bsp = new Bsp(ioService);
}

void BSP_destroy()
{
    delete g_bsp;
}


/*..........................................................................*/
void BSP_SystemReset(void)
{
    //Is this needed?
    //It does not hurt to call ioService.stop() if we are already stopped.
    g_bsp->stop();
}

void BSP_notify()
{
    g_bsp->notify();
}

/*..........................................................................*/
void QF_onCleanup(void) {                               /* cleanup callback */
    printf("\nBye! Bye!\n");
    QS_EXIT();                                    /* perfomr the QS cleanup */
    g_bsp->ioService().stop();
}

void printQueue(QActive *active, QEQueue *queue)
{
    fprintf(stderr, "%12s/%20s min:%3d free:%3d size:%3d events:",
        (active ? active->fileName : ""), queue->name, queue->nMin, queue->nFree, queue->end);
    if (queue->frontEvt) {
        fprintf(stderr, " %d", queue->frontEvt->sig);
    }
    for (unsigned int i = queue->tail; i != queue->head; ) {
        fprintf(stderr, " %d", queue->ring[i]->sig);
        if (i == 0) {
            i = queue->end;
        }
        --i;
    }
    fprintf(stderr, "\n");
}

/*..........................................................................*/
void Q_onAssert(char const Q_ROM * const Q_ROM_VAR file, int line) {
    fprintf(stderr, "Assertion failed in %s, line %d\n", file, line);
    fprintf(stderr, "Queues:\n");
    for (int i = 0; i < QF_MAX_ACTIVE; i++) {
        QActive *active = QF_active_[i];
        if (!active) {
            continue;
        }
        for (int j = 0; j < active->queuesSize; j++) {
            printQueue(active, active->queues[j]);
        }
    }
    print_backtrace();
    abort();
}

/*--------------------------------------------------------------------------*/
#ifdef Q_SPY                                         /* define QS callbacks */

enum QSQHsmTstRecords {
    QS_QDPP_DISPLAY = QS_USER
};

/*..........................................................................*/
uint8_t QS_onStartup(void const *arg) {
    static uint8_t qsBuf[1024];                /* 1K buffer for Quantum Spy */
    char host[64];
    char const *src;
    char *dst;
    uint16_t port = 6601;                                   /* default port */
    struct sockaddr_in servAddr;
    struct hostent *server;

    QS_initBuf(qsBuf, sizeof(qsBuf));

    src = (char const *)arg;
    dst = host;
    while ((*src != '\0') && (*src != ':') && (dst < &host[sizeof(host)])) {
        *dst++ = *src++;
    }
    *dst = '\0';
    if (*src == ':') {
        port = (uint16_t)strtoul(src + 1, NULL, 10);
    }

    l_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);       /* TCP socket */
    if (l_sock == -1){
        printf("Socket cannot be created.\n"
               "socket error 0x%08X.",
               errno);
        return (uint8_t)0;
    }

    server = gethostbyname(host);
    if (server == NULL) {
        printf("QSpy host name cannot be resolved.\n"
               "socket error 0x%08X.",
               errno);
        return (uint8_t)0;
    }
    memset(&servAddr, 0, sizeof(servAddr));
    servAddr.sin_family = AF_INET;
    memcpy(&servAddr.sin_addr, server->h_addr, server->h_length);
    servAddr.sin_port = htons(port);
    if (connect(l_sock, (struct sockaddr *)&servAddr, sizeof(servAddr)) == -1)
    {
        printf("Socket cannot be connected to the QSpy server.\n"
               "socket error 0x%08X.",
               errno);
        QS_onCleanup();
        return (uint8_t)0;
    }

    return (uint8_t)1;                                           /* success */
}
/*..........................................................................*/
void QS_onCleanup(void) {
    if (l_sock != -1) {
        close(l_sock);
    }
}
/*..........................................................................*/
QSTimeCtr QS_onGetTime(void) {           /* invoked within critical section */
    return (QSTimeCtr)clock();
}
/*..........................................................................*/
void QS_onFlush(void) {
    uint16_t nBytes = 1000;
    uint8_t const *block;
    while ((block = QS_getBlock(&nBytes)) != (uint8_t *)0) {
        send(l_sock, (char const *)block, nBytes, 0);
        nBytes = 1000;
    }
}
#endif                                                             /* Q_SPY */
/*--------------------------------------------------------------------------*/

} //extern "C"
