#include "database-migration.h"

#include <string>
#include <map>
#include <set>
#include <boost/noncopyable.hpp>
#include <boost/algorithm/string/join.hpp>
#include <boost/scope_exit.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>
#include <ds/sqlite3/sqlite3.h>
#include "util.h"

#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>
#include <digitalSTROM/dsm-api-v2/dsm-api.h>
#include "datamodel.sql.h"

extern "C" {
// include "datamodel.h" fails. Requires bigger cleanup in headers.
uint64_t Datamodel_groups_array_to_int(uint8_t *groups);
uint64_t Datamodel_sceneLock_array_to_int(const uint8_t *sceneLock, uint8_t sceneLockSize);
} //extern "C"

namespace {

struct DsmApiDeleter {
    void operator()(::DsmApiHandle_t x) {
        DsmApiClose(x);  // TODO(someday): call DsmApiClose in DsmApiCleanup?
        DsmApiCleanup(x);
    }
}; //TODO(someday): move to dsmApi library?
typedef std::unique_ptr<void, DsmApiDeleter> DsmApiPtr;

class Reader : boost::noncopyable {
public:
    Reader(const char* dsmApiAddress)
            : m_database(ds::sqlite3::Database::MEMORY_FILE_NAME), m_nextZoneId(1) {
        m_dsmApiPtr.reset(DsmApiInitialize());
        assert(m_dsmApiPtr);


        // Connect to dsm-api. The server may not be ready yet, give it few retries.
        for (int retry = 0; ; retry++) {
            VDSM_NOTICE("Connecting to %s", dsmApiAddress);
            auto result = DsmApiOpen(m_dsmApiPtr.get(), dsmApiAddress, 0);
            if (result == 0) {
                break;
            }
            auto message = std::string("Failed to open connection address:") + dsmApiAddress
                    + " result:" + boost::lexical_cast<std::string>(result);
            VDSM_ERROR("%s", message.c_str());
            if (retry >= 3) {
                throw std::runtime_error(message);
            }
            sleep(15); // DsmApiOpen does wait internally, but not in all cases.
        }
        VDSM_NOTICE("Connected to %s", dsmApiAddress);
        auto startupDelay = 30;
        VDSM_NOTICE("Waiting %d seconds for vdsms to stabilize.", startupDelay);
        sleep(startupDelay);

        m_database.execRaw(CREATE_DB_V14);
        m_database.execRaw(CONVERT_TO_DB_V15);
        m_database.execRaw(CONVERT_TO_DB_V16);
        m_database.execRaw(CONVERT_TO_DB_V17);
        m_database.execRaw(CONVERT_TO_DB_V18);
        m_database.execRaw(CONVERT_TO_DB_V19);

        ds::sqlite3::Transaction transaction(m_database);
        copyVdcs();
        transaction.commit();
    }

    ds::sqlite3::Database& getDatabase() {
        return m_database;
    }

private:
    DsmApiPtr m_dsmApiPtr;
    ds::sqlite3::Database m_database;

    struct Zone {
        uint16_t zoneId;
        uint8_t localZoneId;
        uint8_t numberOfGroups;
        char name[NAME_LEN];
        Zone() : zoneId(0), localZoneId(0), numberOfGroups(0) { bzero(name, sizeof(name)); }
    };
    int m_nextZoneId;
    std::map<int, Zone> m_zoneIdToZones;
    std::set<std::pair<int, int>> m_zoneSensorKeys;

#define THROW(str) \
        throw std::runtime_error(str \
                + " FILE:" + __FILE__ \
                + " LINE:" + boost::lexical_cast<std::string>(__LINE__));

#define CHECK_RESULT(result) \
    if (result) { \
        THROW(std::string("DsmApi call failed. result:" + boost::lexical_cast<std::string>(result))); \
    }

    uint8_t getLocalZoneId(int zoneId) {
        auto zoneIt = m_zoneIdToZones.find(zoneId);
        if (zoneIt == m_zoneIdToZones.end()) {
            THROW(std::string("Failed to find zoneId:") + boost::lexical_cast<std::string>(zoneId));
        }
        return zoneIt->second.localZoneId;
    }

    void copyVdcs() {
        std::array<dsuid_t, 63> dsmDsuidsArray;
        int dsmDsuidsSize = 0;
        {
            auto result = DsmApiGetBusMembers(
                    m_dsmApiPtr.get(), dsmDsuidsArray.data(), dsmDsuidsArray.size());
            if (result < 0) {
                THROW(std::string("Failed call to DsmApiGetBusMembersPerType result:")
                        + boost::lexical_cast<std::string>(result));
            }
            dsmDsuidsSize = result;
        }
        std::vector<dsuid_t> vdcDsuids;
        for (int i = 0; i < dsmDsuidsSize; i++) {
            const auto& dsmDsuid = dsmDsuidsArray[i];
            try {
                uint8_t type;
                auto result = BusMember_get_type(m_dsmApiPtr.get(), dsmDsuid, &type);
                CHECK_RESULT(result);
                VDSM_NOTICE("Bus member dsuid:%s, type:%d", strdsuid(&dsmDsuid), type);
                if (type == BusMember_vDC) {
                    vdcDsuids.push_back(dsmDsuid);
                }
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to get type of %s: %s", strdsuid(&dsmDsuid), e.what());
            }
        }
        BOOST_FOREACH(const auto& dsmDsuid, vdcDsuids) {
            try {
                VDSM_NOTICE("Reading vdc:%s", strdsuid(&dsmDsuid));
                copyVdc(dsmDsuid);
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to copy vdc %s: %s", strdsuid(&dsmDsuid), e.what());
            }
        }
    }

    void copyVdc(const dsuid_t& dsmDsuid) {
        char dsmDsuidStr[DSUID_STR_LEN];
        dsuid_to_string_upper(&dsmDsuid, dsmDsuidStr);
        VDSM_ENTER("%s", dsmDsuidStr);
        uint32_t hwVersion;
        uint32_t armSwVersion;
        uint32_t dspSwVersion;
        uint16_t apiVersion;
        dsid_t dsid;
        char name[NAME_LEN];
        uint8_t apartmentId;
        uint8_t flags;
        uint16_t autoPingInterval;
        uint16_t onePingInterval;
        uint8_t warnTemperature;
        uint8_t errorTemperature;
        uint8_t upstreamFrequency;
        auto result = dSMInfo(m_dsmApiPtr.get(), dsmDsuid, &hwVersion, &armSwVersion, &dspSwVersion, &apiVersion,
                &dsid, (uint8_t*) name);
        CHECK_RESULT(result);
        result = dSMProperties_get_apartment_id(m_dsmApiPtr.get(), dsmDsuid, &apartmentId);
        CHECK_RESULT(result);
        result = dSMProperties_get_flags(m_dsmApiPtr.get(), dsmDsuid, &flags);
        CHECK_RESULT(result);
        result = dSMProperties_get_ping_intervals(m_dsmApiPtr.get(), dsmDsuid, &autoPingInterval, &onePingInterval);
        CHECK_RESULT(result);
        result = dSMProperties_get_temperature_limits(m_dsmApiPtr.get(), dsmDsuid, &warnTemperature,
                &errorTemperature);
        CHECK_RESULT(result);
        result = dSMProperties_get_upstream_frequency(m_dsmApiPtr.get(), dsmDsuid, &upstreamFrequency);
        CHECK_RESULT(result);
        VDSM_NOTICE("DsmInfo:%s %d %d %d %d %s %d %d %d %d %d %d %d", strdsuid(&dsmDsuid), hwVersion, armSwVersion, dspSwVersion, apiVersion,
                (uint8_t*) name, apartmentId, flags, autoPingInterval,
                onePingInterval, warnTemperature, errorTemperature, upstreamFrequency);

        auto zones = copyZones(dsmDsuid);
        auto defaultZoneId = [&]() -> int {
            if (zones.empty()) {
                VDSM_NOTICE("No zones -> defaultZoneId will be 0");
                return 0;
            }
            return zones[0].zoneId;
        }();

        m_database.exec("INSERT INTO \"dsm_info\" VALUES(?,?,?,?,?,?,?,?,?,?);",
                dsmDsuidStr,
                apartmentId,
                flags,
                autoPingInterval,
                onePingInterval,
                name,
                warnTemperature,
                errorTemperature,
                upstreamFrequency,
                defaultZoneId);

        copyDevices(dsmDsuid);
        copyGroups(dsmDsuid, zones);
        copyHeatings(dsmDsuid, zones);
        copyHeatingOperationModes(dsmDsuid, zones);
        copyScenes();
        copyZoneSensors(dsmDsuid, zones);
        copyClusters(dsmDsuid);
    };

    std::vector<Zone> copyZones(const dsuid_t& dsmDsuid) {
        VDSM_ENTER("%s", strdsuid(&dsmDsuid));
        std::vector<Zone> zones;
        uint8_t count;
        auto result = ZoneCount(m_dsmApiPtr.get(), dsmDsuid, &count);
        CHECK_RESULT(result);
        VDSM_NOTICE("ZoneCount:%d", count);
        zones.reserve(count);
        for (int i = 0; i < count; i++) {
            try {
                zones.push_back(copyZone(dsmDsuid, i));
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to copy zone dsm:%s i:%d what:%s",
                        strdsuid(&dsmDsuid), i, e.what());
            }
        }
        return zones;
    }

    Zone copyZone(const dsuid_t& dsmDsuid, uint8_t index) {
        VDSM_ENTER("%d", index);
        Zone zone;
        auto result = ZoneInfo_by_index(m_dsmApiPtr.get(), dsmDsuid, index,
                &zone.zoneId, &zone.localZoneId, &zone.numberOfGroups, (uint8_t*) zone.name);
        CHECK_RESULT(result);
        // Do not use localZoneId returned in this call but assign our own local localZoneIds.
        // Received localZoneIds may be conflicting if they are read from multiple vdsms.
        auto zoneIt = m_zoneIdToZones.find(zone.zoneId);
        auto localZoneId = (zoneIt != m_zoneIdToZones.end())
            ? zoneIt->second.localZoneId
            : m_nextZoneId++;
        VDSM_NOTICE("ZoneInfo_by_index:%d %d %d %d '%s' -> localZoneId:%d",
                index, zone.zoneId, zone.localZoneId,
                zone.numberOfGroups, zone.name, localZoneId);
        zone.localZoneId = localZoneId; // override with our own local zone id

        // global map of all unique zones
        auto itOk = m_zoneIdToZones.insert(std::make_pair(zone.zoneId, Zone(zone)));
        if (itOk.second) {
            VDSM_INFO("INSERT zone %d %d %s", zone.localZoneId, zone.zoneId, zone.name);
            m_database.exec("INSERT INTO zones VALUES(?,?,?,?);",
                    zone.localZoneId,
                    zone.zoneId,
                    ds::Nullptr(), // zoneTagId - WRITE ONLY, always NULL?
                    zone.name); // seems to be always NULL
        }
        return zone;
    }

    void copyDevices(const dsuid_t& dsmDsuid) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid));
        uint16_t count;
        auto result = ZoneDeviceCount_all(m_dsmApiPtr.get(), dsmDsuid, 0, &count);
        CHECK_RESULT(result);
        VDSM_NOTICE("ZoneDeviceCount_all:%d", count);
        for (int i = 0; i < count; i++) {
            copyDevice(dsmDsuid, i);
        }
    }

    void copyDevice(const dsuid_t& dsmDsuid, uint16_t index) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid), index);
        uint16_t deviceId;
        uint16_t vendorId;
        uint16_t productId;
        uint16_t functionId;
        uint16_t version;
        uint16_t zoneId;
        uint8_t state;
        uint8_t locked;
        uint8_t outputMode;
        uint8_t ltMode;
        uint8_t groups[GROUPS_LEN];
        char name[NAME_LEN];
        dsuid_t dsuid;
        auto result = DeviceInfo_by_index(m_dsmApiPtr.get(), dsmDsuid, 0, index, &deviceId,
                &vendorId, &productId, &functionId, &version, &zoneId, &state, &locked, &outputMode,
                &ltMode, groups, (uint8_t*) name, &dsuid);
        CHECK_RESULT(result);

        char dsmDsuidStr[DSUID_STR_LEN];
        char dsuidStr[DSUID_STR_LEN];
        dsuid_to_string_upper(&dsmDsuid, dsmDsuidStr);
        dsuid_to_string_upper(&dsuid, dsuidStr);
        auto groupsInt = Datamodel_groups_array_to_int(groups);
        VDSM_NOTICE("Device:%s %d %d %d %d %d %d %d %d %d %d %d %d `%s` `%s`",
                dsmDsuidStr, index, deviceId,
                vendorId, productId, functionId, version, zoneId, state, locked, outputMode,
                ltMode, groupsInt, name, dsuidStr);
        m_database.exec("INSERT INTO devices VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?);",
                deviceId,
                dsuidStr,
                vendorId,
                functionId,
                version,
                locked,
                getLocalZoneId(zoneId),
                groupsInt,
                ltMode,
                outputMode,
                name,
                state, // active INTEGER, \ - DEVICE_INFO::State - TODO(verify)
                dsmDsuidStr);

        try {
            copyDeviceBinaryInputs(dsmDsuid, deviceId);
        } catch (const std::exception &e) {
            VDSM_ERROR("Failed to copy device binary inputs dsm:%s deviceId:%d what:%s",
                    strdsuid(&dsmDsuid), deviceId, e.what());
        }
        try {
            copyDeviceButton(dsmDsuid, deviceId);
        } catch (const std::exception &e) {
            VDSM_ERROR("Failed to copy device button dsm:%s deviceId:%d what:%s",
                    strdsuid(&dsmDsuid), deviceId, e.what());
        }
        try {
            copyDeviceOpcs(dsmDsuid, deviceId);
        } catch (const std::exception &e) {
            VDSM_ERROR("Failed to copy device opcs dsm:%s deviceId:%d what:%s",
                    strdsuid(&dsmDsuid), deviceId, e.what());
        }
        try {
            copyDeviceSensors(dsmDsuid, deviceId);
        } catch (const std::exception &e) {
            VDSM_ERROR("Failed to copy device sensors dsm:%s deviceId:%d what:%s",
                    strdsuid(&dsmDsuid), deviceId, e.what());
        }
        try {
            copyDeviceDefaultChannels(dsmDsuid, deviceId);
        } catch (const std::exception &e) {
            VDSM_ERROR("Failed to copy device default channels dsm:%s deviceId:%d what:%s",
                    strdsuid(&dsmDsuid), deviceId, e.what());
        }
    };

    void copyDeviceBinaryInputs(const dsuid_t& dsmDsuid, uint16_t deviceId) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid), deviceId);
        uint8_t count;
        auto result = DeviceBinaryInput_get_count(m_dsmApiPtr.get(), dsmDsuid, deviceId, &count);
        CHECK_RESULT(result);
        VDSM_NOTICE("DeviceBinaryInput_get_count:%d %d", deviceId, count);
        for (int i = 0; i < count; i++) {
            try {
                copyDeviceBinaryInput(dsmDsuid, deviceId, i);
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to copy device binary input dsm:%s deviceId:%d i:%d what:%s",
                        strdsuid(&dsmDsuid), deviceId, i, e.what());
            }

        }
    }

    void copyDeviceBinaryInput(const dsuid_t& dsmDsuid, uint16_t deviceId, uint8_t index) {
        VDSM_ENTER("%s %d %d", strdsuid(&dsmDsuid), deviceId, index);
        uint8_t targetGroupType;
        uint8_t targetGroup;
        uint8_t binaryInputType;
        uint8_t buttonId;
        auto result = DeviceBinaryInput_get_by_index(m_dsmApiPtr.get(), dsmDsuid, deviceId, index,
                &targetGroupType, &targetGroup, &binaryInputType, &buttonId);
        CHECK_RESULT(result);
        VDSM_NOTICE("DeviceBinaryInput:%d %d %d %d %d %d",
                deviceId, index,
                targetGroupType, targetGroup, binaryInputType, buttonId);
        m_database.exec("INSERT INTO deviceBinaryInputs VALUES(?,?,?,?,?,?);",
                deviceId,
                index,
                (targetGroupType << 6) | (targetGroup & 0x3F),
                binaryInputType,
                buttonId,
                ""); // name - write only. Never read or joined.
    };

    void copyDeviceButton(const dsuid_t& dsmDsuid, uint16_t deviceId) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid), deviceId);
        uint8_t buttonId;
        uint8_t groupMembership;
        uint8_t group;
        uint8_t flags;
        uint8_t channelId;
        uint8_t globalGroup;
        auto result = DeviceButtonInfo_by_device(m_dsmApiPtr.get(), dsmDsuid, deviceId,
                &buttonId, &groupMembership, &group, &flags, &channelId, &globalGroup);
        CHECK_RESULT(result);
        VDSM_NOTICE("deviceButton:%d %d %d %d %d %d",
                deviceId, buttonId, groupMembership, group, flags, channelId, globalGroup);
        m_database.exec("INSERT INTO deviceButtons VALUES(?,?,?,?,?,?,?,?);",
                deviceId,
                (buttonId & 0x0f) | ((groupMembership & 0x0f) << 4),
                group, // buttonUserGroup
                // If remote vdsm.db contains in column buttonUserGroup 255,
                // we store to column buttonUserGroup value groupMembership
                // according to logic in Datamodel_DeviceButtonActiveGroup
                flags & 0x01 ? 1 : 0,
                flags & 0x02 ? 1 : 0,
                flags & 0x04 ? 1 : 0,
                channelId,
                globalGroup);
    };

    void copyDeviceOpcs(const dsuid_t& dsmDsuid, uint16_t deviceId) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid), deviceId);
        uint8_t count;
        auto result = DeviceOPCTable_get_count(m_dsmApiPtr.get(), dsmDsuid, deviceId, &count);
        CHECK_RESULT(result);
        VDSM_NOTICE("DeviceOPCTable_get_count:%d %d", deviceId, count);
        for (int i = 0; i < count; i++) {
            try {
                copyDeviceOpc(dsmDsuid, deviceId, i);
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to copy device opc dsm:%s deviceId:%d i:%d what:%s",
                        strdsuid(&dsmDsuid), deviceId, i, e.what());
            }
        }
    }

    void copyDeviceOpc(const dsuid_t& dsmDsuid, uint16_t deviceId, uint8_t index) {
        VDSM_ENTER("%s %d %d", strdsuid(&dsmDsuid), deviceId, index);
        uint8_t channelId;
        auto result = DeviceOPCTable_get_by_index(m_dsmApiPtr.get(), dsmDsuid, deviceId, index,
                &channelId);
        CHECK_RESULT(result);
        VDSM_NOTICE("DeviceOpc:%d %d %d",deviceId, index, channelId);
        m_database.exec("INSERT INTO deviceOpcs VALUES(?,?,?);",
                deviceId,
                index,
                channelId);
    }

    void copyDeviceSensors(const dsuid_t& dsmDsuid, uint16_t deviceId) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid), deviceId);
        uint8_t count;
        auto result = DeviceSensor_get_count(m_dsmApiPtr.get(), dsmDsuid, deviceId, &count);
        CHECK_RESULT(result);
        VDSM_NOTICE("DeviceSensor_get_count:%d %d", deviceId, count);
        for (int i = 0; i < count; i++) {
            try {
                copyDeviceSensor(dsmDsuid, deviceId, i);
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to copy device sensor dsm:%s deviceId:%d i:%d what:%s",
                        strdsuid(&dsmDsuid), deviceId, i, e.what());
            }
        }
    }

    void copyDeviceSensor(const dsuid_t& dsmDsuid, uint16_t deviceId, uint8_t index) {
        VDSM_ENTER("%s %d %d", strdsuid(&dsmDsuid), deviceId, index);
        uint8_t type;
        uint32_t pollInterval;
        uint8_t broadcast;
        uint8_t pushConvert;
        auto result = DeviceSensor_get_by_index(m_dsmApiPtr.get(), dsmDsuid, deviceId, index,
                &type, &pollInterval, &broadcast, &pushConvert);
        CHECK_RESULT(result);
        VDSM_NOTICE("DeviceSensor:%d %d %d %d %d %d %d",deviceId, deviceId, index,
                type, pollInterval, broadcast, pushConvert);
        m_database.exec("INSERT INTO deviceSensors VALUES(?,?,?,?,?,?,?);",
                deviceId,
                index,
                type,
                pollInterval,
                broadcast,
                pushConvert,
                ""); // name - write only. Never read or joined.
    }

    void copyDeviceDefaultChannels(const dsuid_t& dsmDsuid, uint16_t deviceId) {
        // Write only table. No read / join.
    }

    void copyGroups(const dsuid_t& dsmDsuid, const std::vector<Zone>& zones) {
        // TODO(soon):It looks that group table is empty on all vdsm installations.
        // Rather do not populate groups table here. It could lead to regressions.
        return;
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid));
        BOOST_FOREACH(const auto& zone, zones) {
            for (int i = 0; i < zone.numberOfGroups; i++) {
                try {
                    copyGroup(dsmDsuid, zone, i);
                } catch (const std::exception &e) {
                    VDSM_ERROR("Failed to copy group dsm:%s zoneId:%d, i:%d what:%s",
                            strdsuid(&dsmDsuid), zone.zoneId, i, e.what());
                }
            }
        }
    }

    void copyGroup(const dsuid_t& dsmDsuid, const Zone& zone, uint8_t index) {
        auto zoneId = zone.zoneId;
        VDSM_ENTER("%s %d %d %d", strdsuid(&dsmDsuid), zoneId, zone.localZoneId, index);
        uint8_t groupId;
        uint8_t stateMachineGroupId;
        uint16_t numberOfDevices;
        char name[NAME_LEN];
        uint8_t stopTimeout;
        uint8_t remoteStopTimeout;
        auto result = ZoneGroupInfo_by_index(m_dsmApiPtr.get(), dsmDsuid, zoneId, index,
                &groupId, &stateMachineGroupId, &numberOfDevices, (uint8_t*) name, &stopTimeout, &remoteStopTimeout);
        CHECK_RESULT(result);
        VDSM_NOTICE("Group: %d %d %d %d %d %s %d %d", zoneId, index,
                groupId, stateMachineGroupId, numberOfDevices, name, stopTimeout, remoteStopTimeout);
        // first column is primary key, we rely on db to eliminate duplicates
        m_database.exec("INSERT OR IGNORE INTO groups VALUES(?,?,?,?,?,?);",
                zone.localZoneId << 8 | groupId,
                stateMachineGroupId,
                stopTimeout,
                remoteStopTimeout,
                ds::Nullptr(), // TODO(soon): `name` is often corrupted
                0); // stopTimeoutActive - never read or joined
    }

    void copyHeatings(const dsuid_t& dsmDsuid, const std::vector<Zone>& zones) {
        VDSM_ENTER("%s", strdsuid(&dsmDsuid));
        BOOST_FOREACH(const auto& zone, zones) {
            try {
                copyHeating(dsmDsuid, zone.zoneId);
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to copy zone dsm:%s zoneId:%d what:%s",
                        strdsuid(&dsmDsuid), zone.zoneId, e.what());
            }
        }
    }

    void copyHeating(const dsuid_t& dsmDsuid, uint16_t zoneId) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid), zoneId);
        uint8_t controllerMode;
        uint16_t kp;
        uint8_t ts;
        uint16_t ti;
        uint16_t kd;
        uint16_t iMin;
        uint16_t iMax;
        uint8_t yMin;
        uint8_t yMax;
        uint8_t antiWindUp;
        uint8_t keepFloorWarm;
        uint16_t sourceZoneId;
        uint16_t offset;
        uint8_t manualValue;
        uint8_t emergencyValue;
        auto result = ControllerHeating_get_config(m_dsmApiPtr.get(), dsmDsuid, zoneId,
                &controllerMode,
                &kp,
                &ts,
                &ti,
                &kd,
                &iMin,
                &iMax,
                &yMin,
                &yMax,
                &antiWindUp,
                &keepFloorWarm,
                &sourceZoneId,
                &offset,
                &manualValue,
                &emergencyValue);
        CHECK_RESULT(result);
        VDSM_NOTICE("Heating:%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d", zoneId,
                kp,
                kd,
                ti,
                iMin,
                iMax,
                yMin,
                yMax,
                ts,
                antiWindUp,
                sourceZoneId,
                offset,
                emergencyValue,
                manualValue,
                keepFloorWarm,
                controllerMode
        );
        // first column is primary key, we rely on db to eliminate duplicates
        m_database.exec("INSERT OR IGNORE INTO heating VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);",
                zoneId,
                kp,
                kd,
                ti,
                iMin,
                iMax,
                yMin,
                yMax,
                ts,
                antiWindUp,
                sourceZoneId,
                offset,
                emergencyValue,
                manualValue,
                keepFloorWarm,
                controllerMode,
                "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF");
    }

    void copyHeatingOperationModes(const dsuid_t& dsmDsuid, const std::vector<Zone>& zones) {
        VDSM_ENTER("%s", strdsuid(&dsmDsuid));
        BOOST_FOREACH(const auto& zone, zones) {
            try {
                copyHeatingOperationMode(dsmDsuid, zone.zoneId);
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to copy heating operation mode dsm:%s zoneId:%d what:%s",
                        strdsuid(&dsmDsuid), zone.zoneId, e.what());
            }
        }
    }

    void copyHeatingOperationMode(const dsuid_t& dsmDsuid, uint16_t zoneId) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid), zoneId);
        static constexpr int SIZE = 16;
        uint16_t x[SIZE];
        auto result = ControllerHeating_get_operation_modes(m_dsmApiPtr.get(), dsmDsuid, zoneId,
                &x[0], &x[1], &x[2], &x[3], &x[4], &x[5], &x[6], &x[7],
                &x[8], &x[9], &x[10], &x[11], &x[12], &x[13], &x[14], &x[15]);
        CHECK_RESULT(result);
        VDSM_NOTICE("HeatingOperationMode:%d %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x"
                " %02x %02x %02x %02x %02x %02x", zoneId,
                x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7],
                x[8], x[9], x[10], x[11], x[12], x[13], x[14], x[15]);
        auto xEnd = x + SIZE;
        if (xEnd == std::find_if(x, xEnd, [](uint16_t xi) { return xi != 0; })) {
            VDSM_INFO("Skipping empty heating operation mode record");
            return;
        }
        // first column is primary key, we rely on db to eliminate duplicates
        m_database.exec("INSERT OR IGNORE INTO heatingOperationMode VALUES(?,?);",
            zoneId,
            reinterpret_cast<const uint8_t(&)[sizeof(x)]>(x));
    }

    void copyScenes() {
        // Always empty table. No insert ever.
    }

    void copyZoneSensors(const dsuid_t& dsmDsuid, const std::vector<Zone>& zones) {
        VDSM_ENTER("%s", strdsuid(&dsmDsuid));
        BOOST_FOREACH(const auto& zone, zones) {
            const auto& zoneId = zone.zoneId;
            for (int sensorType = 9; sensorType <= 25; sensorType++) {
                try {
                    copyZoneSensor(dsmDsuid, zoneId, sensorType);
                } catch (const std::exception &e) {
                    // We are only guessing sensor types.
                    // We cannot distinguish error from missing value
                    VDSM_NOTICE("No zone sensor or failed to copy zone sensor zoneId:%d sensorType:%d what:%s",
                            zoneId, sensorType, e.what());
                }
            }
        }
    }

    void copyZoneSensor(const dsuid_t& dsmDsuid, uint16_t zoneId, uint8_t sensorType) {
        VDSM_ENTER("%s %d %d", strdsuid(&dsmDsuid), zoneId, sensorType);
        dsuid_t dsuid;
        auto result = ZoneProperties_get_zone_sensor(m_dsmApiPtr.get(), dsmDsuid, zoneId, sensorType, &dsuid);
        CHECK_RESULT(result);
        char dsuidStr[DSUID_STR_LEN];
        dsuid_to_string_upper(&dsuid, dsuidStr);
        VDSM_NOTICE("ZoneSensor: %d %d %s", zoneId, sensorType, dsuidStr);
        auto itOk = m_zoneSensorKeys.insert(std::make_pair(zoneId, sensorType));
        if (itOk.second) {
            VDSM_INFO("INSERT zoneSensor %d", zoneId);
            m_database.exec("INSERT INTO zoneSensors VALUES(?,?,?);",
                    getLocalZoneId(zoneId),
                    sensorType,
                    dsuidStr);
        }
    }

    void copyClusters(const dsuid_t& dsmDsuid) {
        VDSM_ENTER("%s", strdsuid(&dsmDsuid));
        for (int clusterId = 16; clusterId <= 39; clusterId++) {
            try {
                copyCluster(dsmDsuid, clusterId);
            } catch (const std::exception &e) {
                VDSM_ERROR("Failed to copy cluster dsm:%s clusterId:%d what:%s",
                        strdsuid(&dsmDsuid), clusterId, e.what());
            }
        }
    }

    void copyCluster(const dsuid_t& dsmDsuid, uint8_t clusterId) {
        VDSM_ENTER("%s %d", strdsuid(&dsmDsuid), clusterId);
        uint8_t stateMachineGroupId;
        uint8_t hasStateMachine;
        uint16_t numberOfDevices;
        char name[NAME_LEN];
        uint8_t stopTimeout;
        uint8_t remoteStopTimeout;
        uint8_t configurationLock;
        uint8_t sceneLock[16];
        uint16_t location;
        uint16_t floor;
        uint16_t protectionClass;

        auto result = ClusterInfo_by_id(m_dsmApiPtr.get(), dsmDsuid, clusterId,
                &stateMachineGroupId,
                &hasStateMachine,
                &numberOfDevices,
                (uint8_t*) name,
                &stopTimeout,
                &remoteStopTimeout,
                &configurationLock,
                sceneLock,
                &location,
                &floor,
                &protectionClass);
        CHECK_RESULT(result);
        auto sceneLockInt = Datamodel_sceneLock_array_to_int(sceneLock, sizeof(sceneLock));
        auto x = sceneLock;
        VDSM_NOTICE("Cluster: %d %d %d %d %s %d %d %d %d"
                "(%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x) "
                "%d %d %d", clusterId,
                stateMachineGroupId,
                hasStateMachine,
                numberOfDevices,
                name,
                stopTimeout,
                remoteStopTimeout,
                configurationLock,
                sceneLockInt,
                x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7],
                x[8], x[9], x[10], x[11], x[12], x[13], x[14], x[15],
                location,
                floor,
                protectionClass);
        // first column is primary key, we rely on db to eliminate duplicates
        m_database.exec("INSERT OR IGNORE INTO cluster VALUES(?,?,?,?,?,?,?,?,?);",
                clusterId,
                stateMachineGroupId,
                stopTimeout,
                remoteStopTimeout,
                name,
                location,
                floor,
                protectionClass,
                configurationLock);
        m_database.exec("INSERT OR IGNORE INTO sceneLockCluster VALUES(?,?);",
                clusterId,
                sceneLockInt);
    }

#undef CHECK_RESULT
#undef THROW
};

} // namespace

void runDatabaseMigration(const char *databasePath, const char *dsmApiAddress) {
    try {
        ds::sqlite3::Database database(databasePath, ds::sqlite3::Database::Mode::ReadOnly);
        int dbVersion;
        std::tie(dbVersion) = database.exec("SELECT value FROM db_config WHERE key='db_version';").getRow<int>();
        VDSM_NOTICE("Database migration skipped. dbVersion:%d", dbVersion);
        return;
    } catch (const std::exception &e) {
        VDSM_NOTICE("Database not ready, starting migration: %s", e.what());
    }

    VDSM_NOTICE("Migrating vdsm.db over dsm-api %s", dsmApiAddress);
    Reader reader(dsmApiAddress);
    VDSM_NOTICE("Writing database %s", databasePath);
    ds::sqlite3::Database database(databasePath);
    reader.getDatabase().backupTo(database);
    VDSM_NOTICE("Migrating vdsm.db done");
}
