/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <ds/cstddef.h>
#include <memory>
#include <string>

typedef struct sqlite3 sqlite3;
typedef struct sqlite3_stmt sqlite3_stmt;

namespace ds {

/// Lightweight C++ wrapper around sqlite3 C api.
///
/// There are (too) many C++ sqlite3 wrappers. But is there one that is
/// thin, high quality, used on linux and popular at the same time?
///
/// The main motivation for this wrapper to exist and to not use raw C api
/// is exception safety and to propagate errors as exceptions.
/// But it also adds a lot of C++ convenience.

namespace sqlite3 {

class Statement;

/// RAII encapsulatation of sqlite3 database.
/// Statement begins in the constructor and if it was not committed,
/// it is rolled back in destructor.
class Database {
public:
    static const char *MEMORY_FILE_NAME;
    enum class Mode { ReadWrite, ReadOnly };

    /// \brief Initialize database using the specified location.
    ///
    /// If the database does not exist, it will be automatically created.
    Database(const char *fileName, Mode mode = Mode::ReadWrite);

    /// Deleted move constructor because we pass references
    /// to `*this` to other classes for our methods.
    /// But it would be nice to allow copy elision.
    Database(Database &&) = delete;

    /// Call sqlite3_exec with given sql.
    /// Can process multiple statements inside sql. No binding performed.
    void execRaw(const char *sql);
    void execRaw(const std::string &sql) { execRaw(sql.c_str()); }

    /// Create new statement, bind it, call step and return the statement.
    /// Throws on error.
    /// Use `statement.getBusy()` if to determine whether the result set is
    /// non-empty.
    /// Use `statement.getRow<>()` to get first row.
    /// Use `for (auto&& dummy: statement) { statement.getRow<int, int>(); }` to
    /// iterate over results.
    /// Use `for (auto&& row: statement.getRows<int. int>()) { std::get<0>(row);
    /// }` to iterate over results.
    template <typename... Args>
    Statement exec(const char *sql, Args &&... args);
    template <typename... Args>
    Statement exec(const std::string &sql, Args &&... args);

    /// Backup whole database to destination database in one busy operation.
    /// Destination database is completely overwritten.
    void backupTo(::sqlite3 *destination);
    void backupTo(Database &destination) { backupTo(destination.get()); }

    /// Get raw sqlite3* pointer for use in sqlite3 api not wrapped here.
    ::sqlite3 *get() { return m_ptr.get(); }

private:
    struct Deleter {
        void operator()(::sqlite3 *);
    };
    std::unique_ptr< ::sqlite3, Deleter> m_ptr;
};

/// RAII encapsulatation of sqlite3 statement.
/// Statement is prepared in the constructor and finalized in destructor.
class Statement {
public:
    Statement(Database &db, const char *sql);
    Statement(Statement &&) = default; // gcc4.5 does not generate it

    /// @return true if the prepared statement has been stepped at least once
    // but has neither run to completion (step() returned Done)
    // nor been reset using reset().
    bool getBusy() const;
    void assumeBusy(bool busy) const;

    /// Call sqlite3_reset.
    /// Contrary to intuition, it does affect bindings life time.
    /// Bindings must be active during first step() call.
    void reset();

    struct BindDeleter {
        void operator()(::sqlite3_stmt *);
    };
    /// Call sqlite3_clear_bindings after leaving your current scope
    typedef std::unique_ptr< ::sqlite3_stmt, BindDeleter> BindScope;

    /// Bind arguments to statement.
    /// @return BindScope unbinding arguments in its destructor.
    /// All bound strings and blobs must be valid till BindScope is destroyed.
    /// BindScope must be destroyed before another call to bind().
    /// TODO(someday) gcc ignores (warn_unused_result) for templates
    /// https://gcc.gnu.org/bugzilla/show_bug.cgi?id=77542
    template <typename... Args>
    BindScope bind(Args &&... args) __attribute__((warn_unused_result));

    enum class StepResult { Row, Done };

    /// Calls sqlite3_step and returns StepResults. All other return values throw
    /// exception.
    StepResult step();

    /// Reset, bind, and step.
    template <typename... Args>
    StepResult exec(const char *sql, Args &&... args);
    template <typename... Args>
    StepResult exec(const std::string &sql, Args &&... args) {
        return exec(sql.c_str(), std::forward<Args>(args)...);
    }

    /// Get column of current row.
    /// Reference types will be invalidate by non-const methods like step(),
    /// reset(), ...
    /// Explicitly specialized for value types we care about.
    template <typename T>
    T getColumn(int i) const;

    /// Get columns of current row as tuple.
    /// Reference types will be invalidate by non-const methods like step(),
    /// reset(), ...
    template <typename... Ts>
    std::tuple<Ts...> getRow() const;

    /// Input iterator returning (Statement &) from current row advanced by
    /// calling step().
    ///
    /// for (auto &&dummy : statement) { cout << statement->getColumn<int>(1) <<
    /// statement->getColumn<int>(2); }
    class iterator;
    iterator begin();
    iterator end();

    /// Range of rows starting from current row advanced by calling step().
    template <typename... Ts>
    class Rows;

    /// Get range of rows from current row advanced by calling step()
    /// till statement is not busy (usually till step() returns Row).
    ///
    /// for (auto &&row : statement.getRows<int, int>()) { cout <<
    /// std::get<0>(row) << std::get<1>(row) ; }
    template <typename... Ts>
    Rows<Ts...> getRows();

    /// Get raw sqlite3_stmt* pointer for use in sqlite3 api not wrapped here.
    sqlite3_stmt *get() { return m_ptr.get(); }

private:
    struct Deleter {
        void operator()(::sqlite3_stmt *);
    };
    std::unique_ptr<sqlite3_stmt, Deleter> m_ptr;

    void bindRecursive(int) {}

    template <typename T, typename... Args>
    void bindRecursive(int column, T &&value, Args &&... args) {
        bindAt(column, std::forward<T>(value));
        bindRecursive(column + 1, std::forward<Args>(args)...);
    }

    template <typename F>
    void bindAtImpl(const F &f);

    // Explicit bind overrided for types we care about.
    void bindAt(int index, Nullptr);
    void bindAt(int index, int value);
    void bindAt(int index, const char *value);
    void bindAt(int index, const std::string &value);
    template <int N>
    void bindAt(int index, const uint8_t (&value)[N]) {
        bindBlobAt(index, value, N);
    }
    // TODO(someday):  span / array_view type to pass blobs
    //    void bindAt(int index, gsl::span<uint_8t> value) { bindBlobAt(index,
    //    value.data(), value.size()); }
    void bindBlobAt(int index, const uint8_t *, int size);
};

template <>
const char *Statement::getColumn<const char *>(int i) const;
template <>
std::string Statement::getColumn<std::string>(int i) const;
template <>
int Statement::getColumn<int>(int i) const;

/// RAII encapsulatation of sqlite3 transaction.
/// Transaction begins in the constructor and if it was not committed,
/// it is rolled back in destructor.
class Transaction {
public:
    explicit Transaction(Database &database);
    Transaction(const Transaction &) = delete;
    Transaction(Transaction &&) = delete;
    ~Transaction();

    void commit();

private:
    Database &m_database;
    enum class State { Active, Committed };
    State m_state;
};

// implementation details

template <typename... Args>
Statement Database::exec(const char *sql, Args &&... args) {
    Statement statement(*this, sql);
    auto &&bindScope __attribute__((unused)) = statement.bind(std::forward<Args>(args)...);
    statement.step();
    return statement;
}

template <typename... Args>
Statement Database::exec(const std::string &sql, Args &&... args) {
    return exec(sql.c_str(), std::forward<Args>(args)...);
}

template <typename... Args>
Statement::BindScope Statement::bind(Args &&... args) {
    assumeBusy(false);
    BindScope bindScope(get());
    bindRecursive(1, std::forward<Args>(args)...);
    return std::move(bindScope);
}

template <typename... Args>
Statement::StepResult Statement::exec(const char *sql, Args &&... args) {
    reset();
    auto &&bindScope = bind(std::forward<Args>(args)...);
    return step();
}

namespace _ {
template <typename Tuple, int Column, typename... Ts>
struct GetRow;

template <typename Tuple, int Column>
struct GetRow<Tuple, Column> {
    static void f(const Statement &, Tuple &) {}
};
template <typename Tuple, int Column, typename T, typename... Ts>
struct GetRow<Tuple, Column, T, Ts...> {
    static void f(const Statement &statement, Tuple &tuple) {
        std::get<Column>(tuple) = statement.getColumn<T>(Column);
        GetRow<Tuple, Column + 1, Ts...>::f(statement, tuple);
    }
};
};

template <typename... Ts>
std::tuple<Ts...> Statement::getRow() const {
    assumeBusy(true);
    typedef std::tuple<Ts...> Tuple;
    Tuple tuple;
    _::GetRow<Tuple, 0, Ts...>::f(*this, tuple);
    return tuple;
}

class Statement::iterator {
    Statement *m_statement;
    bool getBusy() const { return m_statement ? m_statement->getBusy() : false; }

public:
    iterator() : m_statement(0) {}
    iterator(Statement *statement) : m_statement(statement) {}
    iterator &operator++() {
        m_statement->step();
        return *this;
    }
    iterator operator++(int) {
        auto retval = *this;
        ++(*this);
        return retval;
    }
    bool operator==(iterator other) const { return getBusy() == other.getBusy(); }
    bool operator!=(iterator other) const { return !(*this == other); }
    Statement &operator*() { return *m_statement; }
    typedef std::input_iterator_tag iterator_category;
};
inline Statement::iterator Statement::begin() { return iterator(this); }
inline Statement::iterator Statement::end() { return iterator(); }

template <typename... Ts>
class Statement::Rows {
    Statement *m_statement;

public:
    Rows(Statement *statement) : m_statement(statement) {}
    class iterator {
        Statement *m_statement;
        bool getBusy() const { return m_statement ? m_statement->getBusy() : false; }

    public:
        iterator() : m_statement(0) {}
        iterator(Statement *statement) : m_statement(statement) {}
        iterator &operator++() {
            m_statement->step();
            return *this;
        }
        iterator operator++(int) {
            auto retval = *this;
            ++(*this);
            return retval;
        }
        bool operator==(iterator other) const { return getBusy() == other.getBusy(); }
        bool operator!=(iterator other) const { return !(*this == other); }
        std::tuple<Ts...> operator*() { return m_statement->getRow<Ts...>(); }
        typedef std::input_iterator_tag iterator_category;
    };
    iterator begin() { return iterator(m_statement); }
    iterator end() { return iterator(); }
};

template <typename... Ts>
Statement::Rows<Ts...> Statement::getRows() {
    return Rows<Ts...>(this);
}

} // namespace sqlite3
} // namespace ds
