/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "security.h"
#include "dsm.h"
#include "selectA.h"
#include "registers.h"
#include "bsp.h"
#include "datamodel.h"
#include "button.h"

Q_DEFINE_THIS_MODULE(security)

/* FSM state-handler functions */
static QState Security_initial(Security *me, QEvent const *e);

static QState Security_top(Security *me, QEvent const *e);
static QState Security_idle(Security *me, QEvent const *e);
static QState Security_idle_hold(Security *me, QEvent const *e);
static QState Security_any_alarm(Security *me, QEvent const *e);
static QState Security_any_alarm_hold(Security *me, QEvent const *e);

/* timeout for ending hold */
#define SECURITY_HOLD_END_TIMEOUT_TICKS     (2100/BSP_MS_PER_TICKS)

/* timeout for panic/alarm actions */
#define SECURITY_ACTION_TIMEOUT_TICKS       (1700/BSP_MS_PER_TICKS)


/*..........................................................................*/
void Security_ctor(Security *me)
{
    QHsm_ctor(&me->super, (QStateHandler)&Security_initial);
}

/* HSM definition */

/*..........................................................................*/
static QState Security_initial(Security *me, QEvent const *e)
{
    /* QS dictionary entries */
    QS_FUN_DICTIONARY(&Security_top);
    QS_FUN_DICTIONARY(&Security_idle);
    QS_FUN_DICTIONARY(&Security_idle_hold);
    QS_FUN_DICTIONARY(&Security_any_alarm);
    QS_FUN_DICTIONARY(&Security_any_alarm_hold);

    return Q_TRAN(&Security_idle);
}

/*..........................................................................*/
static QState Security_top(Security *me, QEvent const *e)
{
    switch (e->sig) {
    case CALL_SCENE_BROADCAST_SIG: {
        SceneBasicEvt const * basicEvt = (SceneBasicEvt const *)e;
        uint8_t sceneId = basicEvt->sceneId;
        if (SelectA_SecurityAlarmScene() == sceneId) {
            me->scene = sceneId;
            return Q_TRAN(&Security_any_alarm);
        }
        /* don't change state */
        return Q_HANDLED();
    }
    default: {
        if ((e->sig >= CLICK_MIN_SIG) && (e->sig <= CLICK_MAX_SIG)) {
            /* send general acknowledge for any unhandled click/tip */
            SelectA_SendGeneralAck();
            return Q_HANDLED();
        }
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Security_idle(Security *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_HOLD_START_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case CT_HOLD_REPEAT_SIG: {
        if (Fid_IsSecurityAlarmButton(me->fid, me->key)) {
            me->scene = ALARM1;
            return Q_TRAN(&Security_idle_hold);
        }
        if (Fid_IsSecurityPanicButton(me->fid, me->key)) {
            me->scene = PANIC;
            return Q_TRAN(&Security_idle_hold);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Security_top);
}

/*..........................................................................*/
static QState Security_idle_hold(Security *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, SECURITY_ACTION_TIMEOUT_TICKS, SECURITY_ACTION_SIG, BUTTON_ZONE_0, me->localZoneId, me->groupId, me->deviceId, FALSE, 0);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, SECURITY_HOLD_END_TIMEOUT_TICKS, END_SECURITY_SIG, BUTTON_ZONE_0, me->localZoneId, me->groupId, me->deviceId, FALSE, 0);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, SECURITY_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_SECURITY_SIG: {
        return Q_TRAN(&Security_idle);
    }
    case SECURITY_ACTION_SIG: {
        /* call current alarm scene on all devices in all zones */
        SelectA_CallSceneZoneGroup(0, 0, me->localZoneId, me->groupId, me->scene, me->deviceId);
        return Q_TRAN(&Security_any_alarm);
    }
    }
    return Q_SUPER(&Security_idle);
}

/*..........................................................................*/
static QState Security_any_alarm(Security *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        /* call undo scene on all devices in all zones if scene and button matches */
        bool undoScene = FALSE;
        if (Fid_IsSecurityAlarmButton(me->fid, me->key) && (SelectA_SecurityAlarmScene() == ALARM1)) {
            undoScene = TRUE;
        } else if (Fid_IsSecurityPanicButton(me->fid, me->key)) {
            if (SelectA_SecurityAlarmScene() == PANIC) {
                undoScene = TRUE;
            } else if (SelectA_SecurityAlarmScene() == FIRE) {
                undoScene = TRUE;
                /* send general acknowledge */
                SelectA_SendGeneralAck();
            }
        }
        if (undoScene) {
            SelectA_UndoSceneNumberBroadcast(me->localZoneId, me->groupId, me->scene, me->deviceId);
            return Q_TRAN(&Security_idle);
        }
        /* send general acknowledge from top state */
        break;
    }
    case CT_HOLD_START_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        if (Fid_IsSecurityAlarmButton(me->fid, me->key)) {
            me->scene = ALARM1;
            return Q_TRAN(&Security_any_alarm_hold);
        }
        if (Fid_IsSecurityPanicButton(me->fid, me->key)) {
            me->scene = PANIC;
            return Q_TRAN(&Security_any_alarm_hold);
        }
        return Q_HANDLED();
    }
    case CT_HOLD_REPEAT_SIG: {
        /* do nothing for hold repeat */
        return Q_HANDLED();
    }
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        SceneBasicEvt const * basicEvt = (SceneBasicEvt const *)e;
        uint8_t sceneId = basicEvt->sceneId;
        if (me->scene == sceneId) {
            return Q_TRAN(&Security_idle);
        }
        /* don't change state */
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Security_top);
}

/*..........................................................................*/
static QState Security_any_alarm_hold(Security *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, SECURITY_ACTION_TIMEOUT_TICKS, SECURITY_ACTION_SIG, BUTTON_ZONE_0, me->localZoneId, me->groupId, me->deviceId, FALSE, 0);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, SECURITY_HOLD_END_TIMEOUT_TICKS, END_SECURITY_SIG, BUTTON_ZONE_0, me->localZoneId, me->groupId, me->deviceId, FALSE, 0);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, SECURITY_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_SECURITY_SIG: {
        return Q_TRAN(&Security_any_alarm);
    }
    case SECURITY_ACTION_SIG: {
        /* call alarm scene on all devices in all zones */
        SelectA_CallSceneZoneGroup(0, 0, me->localZoneId, me->groupId, me->scene, me->deviceId);
        /* wait for hold end */
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Security_any_alarm);
}
