/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "VdcFramer.h"

#include <cassert>

#include "util.h"
#include <boost/asio/read.hpp>
#include <boost/asio/write.hpp>

typedef boost::system::error_code error_code;

namespace ds {
namespace vdsm {

VdcFramer::VdcFramer(boost::asio::io_service& ioService)
        : m_ioService(ioService)
        , m_readTimer(ioService)
        , m_socket(0)
        , m_readVector(MaxMessageSize, 0)
        , m_writeStreambuf(MaxWriteBufferSize)
        , m_writeStream(&m_writeStreambuf)
        , m_writeActive(false)
        , m_readTimeout(boost::chrono::seconds(60))
{
    VDSM_IENTER();
}

void VdcFramer::start(boost::asio::ip::tcp::socket& socket)
{
    VDSM_IENTER();

    //Trigger clean up.
    //Method closeImpl() may not have been called yet.
    closeImpl();
    m_socket = &socket;
    asyncReadFrameHeader(); //start async read loop
}

void VdcFramer::close()
{
    //public method
    VDSM_IENTER();
    closeImpl();
}

void VdcFramer::closeImpl()
{
    VDSM_ENTER();
    if (!m_socket) {
        return;
    }
    m_writeStreambuf.consume(m_writeStreambuf.size());
    m_writeActive = false;
    m_socket->close();
    m_readTimer.cancel();
    m_socket = 0;
    m_closed();
}

void VdcFramer::asynWaitReadTimer()
{
    VDSM_ENTER();

    m_readTimer.expires_from_now(m_readTimeout);
    m_readTimer.async_wait([this](const error_code& e)
    {
        if (e) {
            return; //timer was aborted
        }
        auto endpoint = m_socket->remote_endpoint();
        vdsm_report(LOG_WARNING, "Read timeout from %s:%d\n",
            endpoint.address().to_string().c_str(), endpoint.port()
        );
        closeImpl();
    });
}

void VdcFramer::asyncWriteFrame(const void *data, size_t size)
{
    VDSM_ENTER("size:%zd", size);
    if (!m_socket) {
        //TODO this is likely client error.
        return;
    }
    if (m_writeStreambuf.size() + sizeof(FrameHeaderType) + size
        > m_writeStreambuf.max_size()
    ) {
        auto endpoint = m_socket->remote_endpoint();
        vdsm_report(LOG_WARNING, "Write buffer exceeded to %s:%d\n",
            endpoint.address().to_string().c_str(), endpoint.port()
        );
        closeImpl();
        return;
    }
    FrameHeaderType header = htons(size);
    m_writeStream.write(reinterpret_cast<const char *>(&header), sizeof(header));
    m_writeStream.write(reinterpret_cast<const char *>(data), size);
    asyncWrite();
}

void VdcFramer::asyncReadFrameHeader()
{
    VDSM_ENTER();
    auto buffer = boost::asio::buffer((void *)(&m_readHeaderData), sizeof(m_readHeaderData));
    boost::asio::async_read(*m_socket, buffer, [=](const error_code& e, std::size_t n)
    {
        if (e == boost::asio::error::operation_aborted) {
            return; //'this' may already be deleted
        }
        VDSM_ENTER("e:%s", e.message().c_str());
        if (e) {
            closeImpl();
            return;
        }
        assert(n == sizeof(FrameHeaderType)); //async_read promise
        auto headerSize = ntohs(m_readHeaderData);
        if (headerSize <= 0 || headerSize > MaxMessageSize) {
            auto endpoint = m_socket->remote_endpoint();
            vdsm_report(LOG_WARNING, "Invalid read frame size from %s:%d headerSize:%d\n",
                endpoint.address().to_string().c_str(), endpoint.port(), headerSize
            );
            closeImpl();
            return;
        }
        asyncReadFrameBody(headerSize);
    });
    asynWaitReadTimer(); //read timeout guard, we have new frame
}

void VdcFramer::asyncReadFrameBody(FrameHeaderType header)
{
    VDSM_ENTER("header:%d", header);
    auto buffer = boost::asio::buffer(m_readVector.data(), header);
    boost::asio::async_read(*m_socket, buffer,
        [=](const error_code& e, std::size_t n)
    {
        if (e == boost::asio::error::operation_aborted) {
            return; //'this' may already be deleted
        }
        VDSM_ENTER("e:%s", e.message().c_str());
        if (e) {
            closeImpl();
            return;
        }
        assert(n == header); //async_read promise
        assert(n <= m_readVector.size()); //checked in asyncReadFrameHeader

        //Start another read cycle, it may be cancelled in message handler.
        asyncReadFrameHeader();

        m_readFrameFunction(m_readVector.data(), header);
    });
}

void VdcFramer::asyncWrite()
{
    VDSM_ENTER("m_writeActive:%d, m_writeStreambuf.size():%zd", m_writeActive,
        m_writeStreambuf.size()
    );
    if (m_writeActive || m_writeStreambuf.size() == 0) {
        return;
    }
    m_writeActive = true;
    boost::asio::async_write(*m_socket, m_writeStreambuf,
        [=](const error_code& e, std::size_t n)
    {
        if (e == boost::asio::error::operation_aborted) {
            return;
        }
        VDSM_ENTER("e:%s", e.message().c_str());
        if (e) {
            closeImpl();
            return ;
        }
        m_writeActive = false;
        asyncWrite();
    });
}

}} //namespace ds::vdsm
